#!/usr/bin/env nimcr
#nimcr-args c --threads:on -d:release -d:log
include "dependencies/babel.nim"

import hashes
import tables
import threadpool
import sets
import math

let
  dep = getAppDir()/"dependencies"
  aigtoaig = dep/"aiger/aigtoaig "
  aigmove = dep/"aiger/aigmove -i "
  aigor = dep/"aiger/aigor "
  aigtrav = dep/"aigtrav/aigtrav "
  mcaiger = dep/"mcaiger/mcaiger/mcaiger -n "
  camical = dep/"camical/camical -n "
  k_witness_generator = dep/"/certifaiger_pp/aigcertify_kind "
  invariant_checker = dep/"certifaiger_pp/aigcertify/aigcertify -o "
  mapping_check = dep/"certifaiger_pp/mapping_check/mapping_check -o "
  kissat =   (when defined release: dep/"kissat/build/kissat -q "
              else: dep/"kissat/build/kissat -q --relaxed ")
  aigtocnf = (when defined release: dep/"aiger/aigtocnf "
              else: dep/"aiger/aigtocnf_nr  -m --no-pg ")

var tmpFiles: seq[string]
proc tmp(ext: string): string =
  result = getTempDir()/uuid()&"."&ext
  tmpFiles.add result
proc deleteTmpFiles =
  for f in tmpFiles: smd &"rm -f {f}"
  tmpFiles.setLen 0

proc td_certify(file: string) =
  start_logging
  defer: deleteTmpFiles()

  type Val = enum
    F = 0'i8  #False
    T = 1'i8  #True
    X = 2'i8  #Undefined
    I = 3'i8  #Invalid
  var # one circuit in memory at all time
    var_len, state_len, lit_len, mask = -1
    input, output, bad, constraint, justice, justice_constraints, fairness: seq[int]
    gate: seq[tuple[v: int, l: int, r: int]]
    latch: seq[tuple[v: int, n: int, r: int]]
  proc new_var: int = inc var_len; 2*var_len

  # AIGER utils
  func neg(v:int): int = v xor 1
  func neg(v:Val): Val = [T,F,X,I][int v]
  proc conj(l,r: int): int =
    result = new_var()
    gate.add (result, l, r)
  proc disj(l,r: int): int = neg conj(neg l, neg r)
  proc xDisj(l,r: int): int = disj(conj(l, neg r), conj(neg l, r))
  proc eq(l,r: int): int = conj(neg conj(neg l, r), neg conj(l, neg r))
  proc impl(l,r: int): int = disj(neg l, r)
  proc ite(i, t, e: int): int = disj(conj(i,t), conj(neg i, e))
  proc reduce(c: seq[int], f: (int, int) -> int): int =
    var c = c
    while c.len > 1:
      for i in countup(0, c.high - 1, 2): c[i div 2] = f(c[i], c[i+1])
      if c.len mod 2 == 1: c[(c.len+1) div 2 - 1] = c[^1]
      c.setLen (c.len+1) div 2
    return c[^1]
  proc conj(c: seq[int]): int = (if c.len == 0: 1 else: c.reduce conj)
  proc disj(c: seq[int]): int = (if c.len == 0: 0 else: c.reduce disj)
  func pol(l:int): Val = (if (l and 1) == 0: T else: F)
  func ind(v: int): int = assert v >= 2; v.div(2) - 1
  func lit(ind: int, pol = T): int =
    assert pol in [F,T]
    result = (ind+1)*2
    if pol == F: return neg result
  func get[R](s: seq[R], v: int): R =
    if   v < 2: R v
    elif v.pol == T: s[v.ind]
    else: neg s[v.ind]
  func set[R](s: var seq[R], v: int): var R =
    assert v.pol == T
    s[v.ind]

  proc read_aag(file: string) =
    say "read "&file
    let f = open file
    defer: close f
    var head = f.readLine.split[1..^1].map parseInt
    head &= 0.repeat(9 - head.len)
    (M, I, L, O, A, B, C, J, F) := head
    mask = 1.shl((64/M.float.log(10)).int)-1
    if L == 0: echo "Combinatorial circuit not supported"; quit 0
    if O < 1: echo "No property. Exiting."; quit 0
    assert O == 1
    if J + F > 0: echo "Justice and Fairness not supported"; quit 0
    if A == 0: echo "Mapping generator is buggy if there are no gates"; quit 0 # TODO
    var_len = M
    state_len = I + L # size of a state
    lit_len = lit(state_len, Val.F)
    proc read_n[T](collection: var seq[T], n: int) =
      collection = newSeqOfCap[T] n
      for _ in 1..n: collection.add parseInt readLine f
    input.read_n I
    latch = newSeqOfCap[(int,int,int)] L
    for _ in 1..L:
      let l = f.readLine.split.map parseInt
      let r = (if l.len > 2: l[2] else: 0)
      latch.add (l[0], l[1], r)
    output.read_n O
    bad.read_n B
    constraint.read_n C
    justice.read_n J
    if J > 0: justice_constraints.read_n justice.foldl(a+b)
    fairness.read_n F
    gate = newSeqOfCap[(int,int,int)] A
    for _ in 1..A:
      let g = f.readLine.split.map parseInt
      gate.add (g[0], g[1], g[2])

  proc circuite_size: string =
    let lengths = (input, latch, output, gate, bad, constraint, justice, fairness)
    result = $var_len
    for v in fields lengths:
      result &= " " & $len(v)

  proc circuite_size(file: string): string =
    var head = file.lines[0].split[1..^1].map parseInt
    head &= 0.repeat(9 - head.len)
    return head.join " "

  proc write_aag(file: string) =
    say &"write {file}"
    let f = file.open fmWrite
    defer: close f
    func `$`(l: (int,int,int)): string = &"{l[0]} {l[1]} {l[2]}"
    f.write &"aag {var_len}"
    for v in fields (input, latch, output, gate, bad, constraint, justice, fairness):
      f.write " " & $len(v)
    f.write "\n"
    for v in fields (input, latch, output, bad, constraint, justice, justice_constraints, fairness, gate):
      for l in v:
        f.writeLine l

  proc normalize(file: string): string =
    time_block "normalize"
    var
      aag = file
      moved = tmp "aag"
      ored  = tmp "aag"
      aig   = tmp "aig"
    result  = tmp "aag"
    if "only outputs founds" notin
       cmd &"{aigmove} {aag} {moved}":
      aag = moved
    smd &"{aigor} {aag} {ored}"
    smd &"{aigtoaig} {ored} {aig}"
    smd &"{aigtoaig} {aig} {result}"

  proc reencode: Table[int, int] =
    result = {0:0, 1:1}.toTable
    var i = 0
    for v in input:
      i += 2; result[v] = i; result[neg v] = neg i
    for (v,_,_) in latch:
      i += 2; result[v] = i; result[neg v] = neg i
    for (v,_,_) in gate:
      i += 2; result[v] = i; result[neg v] = neg i

    for i, v in input:
      input[i] = result[v]
    for i, (v, n, r) in latch:
      latch[i] = (result[v], result[n], result[r])
    for i, v in output:
      output[i] = result[v]
    for i, (v, l, r) in gate:
      gate[i] = (result[v], result[l], result[r])

  proc check_certificate(model, certificate: string) =
    say &"Check certificate {certificate} for model {model}"
    time_block "check_certificate"
    proc check(n,f: string) =
      assert n.hash != 0
      say &"{n} check {f}"
      time_block n
      let cnf = tmp "cnf"
      smd &"{aigtocnf} {f} {cnf}"
      let r = cmd &"{kissat} {cnf}"
      if "UNSATISFIABLE" notin r:
        if "SATISFIABLE" notin r: echo r
        echo &"\nFailed {n} check {f} -> {cnf}"
        if r.len > 10000: quit hash n
        let
          f = lines f
          b = f[0].split[2].parseInt
          g = f[1..b]&f[(b+2)..^1]
          v = r.split("s SATISFIABLE")[1].strip.replace("\nv")[2..^3].split[1..^1]
               .mapIt(if '-' in it: "0" else: "1")
        for (v, g) in zip(v, g): echo &"{v} | {g}"
        quit hash n
    let f = collect:
      for _ in 1..6: tmp "aag"
    block:
      time_block "inductive"
      smd &"{invariant_checker} {f[0]} {f[1]} {f[2]} {certificate} {certificate}"
      "initiation".check f[0]
      "consistency".check f[1]
      "consecution".check f[2]
    block:
      time_block "mapping"
      smd &"{mapping_check} {f[3]} {f[4]} {f[5]} {model} {certificate}"
      "simf".check f[3]
      "prop".check f[4]
      "reset".check f[5]

  proc k_witness(k: int, file: string): string =
    time_block "k_witness_generation"
    let k_witness = tmp "aag"; show k_witness
    smd &"{k_witness_generator} -k {k} -o {k_witness} {file}"
    if not defined release:
      say "checking k_witness"
      check_certificate(file, k_witness)
    return k_witness

  proc kind(file: string): (int, int) =
    time_block "kind"
    let r = cmd mcaiger & file
    result[0] = r.sub("d=(.*)", "-1").parseInt
    result[1] = r.sub("k=(.*)", "-1").parseInt
    assert result[0] > -1 or result[1] > -1

  proc og_certify(model: string): int =
    time_block "og_certify"
    let (og_d, og_k) = kind model
    if og_k > -1: show og_k
    else: show og_d; return
    var og_certificate = model
    if og_k > 1: og_certificate = k_witness(og_k, model)
    say "og_certificate_size=" & circuite_size og_certificate
    check_certificate(model, og_certificate)
    return og_k

  type Unsafe = object of Exception

  proc bdd_witness =
    time_block "bdd"
    let bdd_res = splitLines strip cmd aigtrav & file
    if bdd_res[^1].parseBool: raise Unsafe.newException "Found bad state in suffix"
    if bdd_res[0] == "1": return # inductive property
    let tokens = split bdd_res[0]
    var lit_table: Table[int, int]
    for t in bdd_res[1..^2]:
      let t = t.split(" -> ").map parseInt
      lit_table[t[0]] = t[1]
    proc lit(v: string): int =
      if v[0] == '!': neg lit_table[parseInt v[2..^1]]
      else: lit_table[parseInt v[1..^1]]
    var
      op: seq[char]
      lits: seq[int]
    for t in tokens:
      case t:
        of "?", "&", "|", "^": op.add t[0]
        of ":": assert op[^1] == '?'
        of "(": continue
        of ")":
          let op = pop op
          if op == '?':
            let e = pop lits
            let t = pop lits
            let i = pop lits
            lits.add ite(i, t, e)
          else:
            let a = pop lits
            let b = pop lits
            if   op == '&': lits.add conj(a, b)
            elif op == '|': lits.add disj(a, b)
            elif op == '^': lits.add xDisj(a, b)
        else: lits.add lit t
    assert op.len == 0 and lits.len == 1
    output[0] = neg lits[0]
    let bdd_witness = tmp "aag"; show bdd_witness
    write_aag bdd_witness

  proc prefix(model: string, bound :int) =
    time_block "prefix"
    if bound < 0: return
    let d_pre = (&"{camical} {bound} {model}").cmd.sub("bad state reached at bound (.*)", "-1").parseInt
    if d_pre > -1:
      show d_pre
      raise Unsafe.newException "Found bad state in prefix"

  proc suffix(forward: string): int =
    time_block "suffix"
    let (d, k) = kind forward
    if d > -1:
      show d
      raise Unsafe.newException "Found bad state in suffix"
    show k
    assert k > -1
    return k

  proc og_bdd(model: string) =
    time_block "og_bdd"
    bdd_witness()
    let og_bdd_certificate = tmp "aig"; show og_bdd_certificate
    write_aag og_bdd_certificate
    say "og_bdd_certificate_size=" & circuite_size og_bdd_certificate
    check_certificate(model, og_bdd_certificate)

  proc model_checking(model, forward: string, duration: int): int =
    result = 1
    time_block "model_checking"
    prefix(model, duration-1)
    when defined bdd: bdd_witness()
    else: return suffix(forward)

  template DEBUG_partial_vs_full =
    when defined matching:
      say "Compare partial and complete cube matching"
      when defined partial: quit "Matching check and partial flag incompatible."
      let (p_trail, p_stem_length) = ternary_simulation[true]()
      let (p_duration, p_transients) = transient_signals(p_trail, p_stem_length)
      let p_cycle_length = p_trail.len - p_stem_length
      show p_stem_length, p_duration, p_transients.len, p_cycle_length
      assert trail.len >= p_trail.len
      assert trail[0..p_trail.high] == p_trail
      assert stem_length >= p_stem_length
      assert transients.len >= p_transients.len
      assert transients.toHashSet >= p_transients.toHashSet

  when defined x:
    var number_transitions, x_max = 0
    var xed_variables: HashSet[int]
  proc transition(state: var seq[Val]) =
    assert state.len == state_len
    var extended = state & I.repeat(var_len - state_len)
    for (v, l, r) in gate:
      extended.set(v) = Val min(X.int, extended.get(l).int * extended.get(r).int)
    when not defined x:
      for (v, n, _) in latch:
        state.set(v) = extended.get n
    else:
      inc number_transitions
      if (number_transitions and mask) != 0:
        for (v, n, _) in latch:
          state.set(v) = extended.get n
      else:
        var x_out = true
        let x_count = latch.countIt extended.get(it.n) == X
        if x_count > x_max:
          x_max = x_count
          x_out = false
        for (v, n, _) in latch:
          var
            old_value = state.get v
            new_value = extended.get n
          if x_out and
            # 0 -> 1 or 1 -> 0
            (old_value.int xor new_value.int) == T.int and
            not xed_variables.containsOrIncl v:
            x_out = false
            say &"xed out {v} during transition {number_transitions} was {old_value} -> {new_value}"
            new_value = X
          state.set(v) = new_value

  proc ternary_simulation[forward_subsumption: static bool]: (seq[seq[int]], int) =
    func cube(s: seq[Val]): seq[int] =
      assert s.len == state_len
      for v, p in s:
        if p == X: continue
        assert lit(v,p) <= lit(state_len, F)
        result.add lit(v,p)
    var
      state = I.repeat state_len
      trail: seq[seq[int]]
    when forward_subsumption:
      var watch = newSeq[int]().repeat lit_len
      # empty cube cannot be watched but subsumes every other cube
      var empty_cube_index = -1
      proc trail_add(c: seq[int]) =
        trail.add c
        if c.len > 0: watch[c[0]].add trail.high
        else: empty_cube_index = trail.high
      proc subsumed(s: seq[Val], c: seq[int]): int =
        if empty_cube_index > -1: return empty_cube_index
        var subsuming_cubes: seq[int]
        result = -1
        for l, w in watch:
          if w.len == 0: continue
        for l in c:
          for index in watch[l]:
            let other_clause = trail[index]
            var unmarked = -1
            for ol in other_clause:
              if s[ol.ind] != ol.pol:
                unmarked = ol
                break
            if unmarked == -1: subsuming_cubes.add index
            else: watch[unmarked].add index
          # If we find a subsumption the ternary simulation terminates.
          # Otherwise the watches are cleared.
          watch[l].setLen 0
        if subsuming_cubes.len > 0: return max subsuming_cubes
    else: # hashing
      var index: Table[Hash, int]
      proc trail_add(c: seq[int]) =
        trail.add c
        index[hash c] = high trail
      proc subsumed(s: seq[Val], c: seq[int]): int =
        result = index.getOrDefault(c.hash,-1)
        assert result == -1 or trail[result] == c
    for i in input: state.set(i) = X
    for (v, n, r) in latch:
      state.set(v) = (if r == 0: F
                      elif r == 1: T
                      elif r == v: X
                      else: I)
      assert state.get(v) != I
    trail_add cube state
    while true:
      transition state
      let c = cube state
      let index = subsumed(state, c)
      if index > -1: return (trail, index)
      trail_add c

  proc transient_signals(trail: seq[seq[int]], stem_length: int): (int, seq[(int, int)]) =
    var count = 0.repeat lit_len
    for c in trail[stem_length..^1]:
      for l in c: inc count[l]
    let maximum = trail.len - stem_length
    var stabilization = X.repeat state_len
    for l, c in count:
      if c < maximum: continue
      stabilization[l.ind] = l.pol
      result[1].add (l.ind-input.len, l.pol.int)
    result[0] = 0
    for i in count_down(stem_length-1, 0):
      var hit = deepCopy stabilization
      for l in trail[i]:
        if hit[l.ind] == l.pol.neg:
          result[0] = i+1; return
        else: hit[l.ind] = X
      if hit.any (s => s != X):
        result[0] = i+1; return

  proc forward_circuit(duration: int, transients: seq[(int, int)]): (seq[seq[int]], seq[(int, int)], seq[seq[int]]) =
    proc pseudo_input: int =
      result = new_var()
      latch.add (result, result, result)
    let (high_latch, high_gate) = (high latch, high gate)
    var
      U = (-1).repeat(var_len).repeat(duration)
      uninitialized_latches: seq[(int,int)]
      pseudo_inputs = (-1).repeat(len input).repeat(duration)
    for j in 0..<duration:
      for i, v in input:
        U[j].set(v) = pseudo_input()
        pseudo_inputs[j][i] = high latch
      for i, (v, n, r) in latch[0..high_latch]:
        U[j].set(v) = (if j > 0: U[j-1].get(n) elif r == v: pseudo_input() else: r)
        if r == v: uninitialized_latches.add (i, high latch)
      for (v, l, r) in gate[0..high_gate]:
        U[j].set(v) = conj(U[j].get(l), U[j].get(r))
    if duration > 0:
      for i, (_, n, _) in latch[0..high_latch]:
        latch[i].r = U[^1].get(n)

    var transient_table: Table[int, int]
    proc to_transient(v: int): int =
      transient_table.getOrDefault(v, v)
    for (i, p) in transients:
      latch[i].r = p
      latch[i].n = p
      transient_table[latch[i].v] = p
      transient_table[latch[i].v.neg] = p.neg
    for l in mitems latch:
      l[1] = to_transient l[1]
      l[2] = to_transient l[2]
    for o in mitems output:
      o = to_transient o
    for g in mitems gate:
      g[1] = to_transient g[1]
      g[2] = to_transient g[2]

    return (U, uninitialized_latches, pseudo_inputs)

  proc temporal_decomposition: (int, seq[(int, int)], seq[seq[int]]) =
    time_block "temporal_decomposition"
    when defined partial: say "Using forward subsumption"
    let (trail, stem_length) = ternary_simulation[defined partial]()
    let (duration, transients) = transient_signals(trail, stem_length)
    let cycle_length = trail.len - stem_length
    show stem_length, duration, transients.len, cycle_length
    DEBUG_partial_vs_full
    return (duration, transients, trail[duration..^1])

  proc compose_witness(stable_cubes: seq[seq[int]]) =
    time_block "compose_witness"
    let cube_loop = disj stable_cubes.map(conj)
    output[0] = output[0].disj neg cube_loop
    say "composit_witness_size=" & circuite_size()

  proc backward_witness(duration, c0_P: int, c0_latch: seq[tuple[v: int, n: int, r: int]], freez: Slice[int], U: seq[seq[int]], uninitialized_latches: seq[(int, int)], pseudo_inputs: seq[seq[int]]) =
    time_block "backward_witness"
    # add bs
    var b = collect:
      for _ in 0..<duration: new_var()
    latch.add (b[0], 1, 0)
    for i in 1..<duration:
      latch.add (b[i], b[i-1], 0)
    # pseudo_inputs copy inputs
    # This works because the PIs are not actual latches and therefore not checked in R.
    # In the next step when they influence the actual latch-values, they copied the correct values from I.
    let m0 = neg b[0]
    for (og, dup) in uninitialized_latches:
      latch[dup].n = ite(m0, latch[og].v, latch[dup].v)
    for i, pi in pseudo_inputs:
      let m = (if i == 0: m0
               else: conj(neg b[i], b[i-1]))
      for  j, l in pi:
        latch[l].n = ite(m, input[j], latch[l].v)
    # freeze with last b
    for i in freez:
      latch[i].n = ite(b[^1], latch[i].n, latch[i].v)
    # Reset property R and V
    var R = (-1).repeat(duration)
    for i in countdown(R.high, 0):
      let c0_latch = (if i>0: c0_latch
                   else: c0_latch.filter(l => l.r != l.v))
      R[i] = conj c0_latch.map(l => eq(l.v, U[i].get(l.v)))
    var V = (-1).repeat(duration)
    V[^1] = conj latch[freez].map(l => eq(l.v, l.r))
    for i in countdown(V.high-1, 0):
      V[i] = V[i+1].conj eq(b[i+1],0)
    # property
    var p = neg output[0]
    for i in countdown(duration-1, 0):
      var q = 1.repeat 4
      q[0] = (neg b[i]).impl c0_P
      q[1] = b[i].impl p
      q[2] = (neg b[i]).impl R[i]
      q[3] = (neg b[i]).impl V[i]
      p = conj q
    output[0] = neg p
    say "backward_witness_size=" & circuite_size()

  show file
  let model = normalize file
  read_aag model
  var original_k = high int
  when (defined og) or (defined ok):
    original_k = og_certify model
    when defined og: quit 0
  when defined og_bdd:
    try: og_bdd model
    except Unsafe as e: say e.msg; quit 0
    quit 0
  say "model_size=" & circuite_size model

  var (duration, transients, stable_cubes) =
    temporal_decomposition()
  when defined td: quit 0

  var c0_latch = deepCopy latch
  var c0_gate = deepCopy gate
  var c0_P = neg output[0]
  var (U, uninitialized_latches, pseudo_inputs) =
    forward_circuit(duration, transients)
  var var_table = reencode()
  let forward = tmp "aag"; show forward
  write_aag forward
  say "forward_size=" & circuite_size forward

  let (first_wit, max_latch) = (len latch, latch[^1].v.neg)
  var k: int
  try: k = model_checking(model, forward, duration)
  except Unsafe as e: say e.msg; quit 0
  assert k <= original_k
  if k > 1: read_aag k_witness(k, forward)
  say "forward_witness_size=" & circuite_size()

  let freez = first_wit..latch.high
  # update stored var values
  for k, v in var_table:
    if v <= max_latch: continue # const, input, latch
    var_table[k] = v + 2*freez.len
  for u in mitems U: u.applyIt var_table[it]
  for l in mitems c0_latch:
    l = (var_table[l.v], var_table[l.n], var_table[l.r])
  for g in mitems c0_gate:
    g = (var_table[g.v], var_table[g.l], var_table[g.r])
  c0_P = var_table[c0_P]

  for i, l in c0_latch: latch[i] = l
  for i, g in c0_gate: gate[i] = g
  if k == 0: output[0] = neg c0_P
  if transients.len > 0:
    compose_witness(stable_cubes)
  if duration > 0:
    backward_witness(duration, c0_P, c0_latch, freez, U, uninitialized_latches, pseudo_inputs)
  var_table = reencode()
  let certificate = (if paramCount()>1: paramStr 2 else: tmp "aag")
  show certificate
  write_aag certificate

  check_certificate(model, certificate)
  say "certifaiger_td produced certificate " & certificate

if paramCount() == 0:
  fuzz:
    let aigfuzz = getAppDir()/"dependencies/aiger/aigfuzz -s -m -1 -a -o"
    let aigdd = getAppDir()/"dependencies/aiger/aigdd "
    let model = "/tmp"/uuid()&".aag"
    lmd &"{aigfuzz} {model}"
    if thisFails model:
      let bug = "bug_"&model.split("/")[^1]
      let dd = "dd_"&model.split("/")[^1]
      smd &"mv {model} {bug}"
      echo bug
      lmd &"{aigdd} {bug} {dd} {paramStr 0}"
      echo dd
    smd &"rm -f {model}"
elif 1.paramStr[^3..^1] in ["aig", "aag"]:
  td_certify paramStr 1
else: echo &"Usage: {paramStr 0} <model.aag/aig> [<certificate.aag>]"
