#include "../aiger/aiger.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/unistd.h>

typedef struct AIG AIG;
typedef struct BDD BDD;
typedef struct Line Line;

enum Op
{
  EMPTY = 0,
  ITE = 1,
  XOR = 2,
  OR = 3,
  AND = 4,
  EX = 5,
  EY = 6,
};

typedef enum Op Op;

struct AIG
{
  unsigned input:1;
  unsigned latch:1;
  unsigned freed:1;
  unsigned and:1;
  unsigned mark:1;
  unsigned valid:1;
  unsigned included:1;
  unsigned coi:1;
  unsigned level;
  unsigned var;
  unsigned ref;
  unsigned rhs[2];
  BDD *bdd;
  BDD *next;
};

struct BDD
{
  unsigned var;
  unsigned idx;
  unsigned ref;
  unsigned mark : 1;
  BDD *next;
  BDD *low;
  BDD *high;
};

struct Line
{
  Op op;
  BDD *a;
  BDD *b;
  BDD *r;
};

static aiger *aiger_model;

static size_t current_bytes;
static size_t max_bytes;

static BDD **bdds;
static unsigned num_bdds;
static unsigned size_bdds;

static unsigned long long idx_bdds;

static BDD **dead_bdds;
static unsigned size_dead_bdds;
static unsigned num_dead_bdds;
static unsigned num_dead_bdds_limit = 10000;

static BDD *recycled_bdds;
static unsigned num_recycled_bdds;

static Line *cache;
static unsigned size_cache;
static unsigned size_cache_limit = (1u << 29) / sizeof (Line); /*512MB*/
static unsigned count_cache;

static struct AIG *aigs;

static unsigned long long unique_table_lookups;
static unsigned long long unique_table_visits;

static unsigned long long cache_lookups;
static unsigned long long cache_hits;

static unsigned ands;
static unsigned xors;
static unsigned ors;
static unsigned eys;

static unsigned long long rands;
static unsigned long long rxors;
static unsigned long long reys;

static unsigned long long apply_steps;
static unsigned long long apply_steps_limit = ULONG_MAX;

static unsigned verbose;
static int last_line_was_empty;

static unsigned * unsigned_stack;
static unsigned size_unsigned_stack;
static unsigned count_unsigned_stack;

struct 
{
  struct {
    unsigned inputs;
    unsigned latches;
    unsigned aigs;
  } coi;

  struct {
    unsigned count;
    unsigned inputs;
    unsigned latches;
    unsigned freed;
    unsigned aigs;
  } abstraction;
} stats;

#define INCBYTES(b) \
do { \
  current_bytes += (b); \
  if (current_bytes > max_bytes) \
    max_bytes = current_bytes; \
} while (0)

#define DECBYTES(b) \
do { \
  assert (current_bytes >= (b)); \
  current_bytes -= (b); \
} while (0)

#define NEWN(p,n) \
do { \
  size_t newn_bytes = sizeof (*(p)) * (n); \
  (p) = malloc (newn_bytes); \
  if (!(p)) \
    die ("out of memory in 'malloc'"); \
  INCBYTES (newn_bytes); \
  memset ((p), 0, newn_bytes); \
} while (0)

#define REALLOCN(p,o,n) \
do { \
  size_t reallocn_old_bytes = sizeof (*(p)) * (o); \
  size_t reallocn_new_bytes = sizeof (*(p)) * (n); \
  DECBYTES (reallocn_old_bytes); \
  (p) = realloc ((p), reallocn_new_bytes); \
  if (!(p)) \
    die ("out of memory in 'realloc'"); \
  INCBYTES (reallocn_new_bytes); \
} while (0)

#define CLRN(p,n) \
do { \
  size_t clrn_bytes = sizeof (*(p)) * (n); \
  memset ((p), 0, clrn_bytes); \
} while (0)

#define DELETEN(p,n) \
do { \
  size_t deleten_bytes = sizeof (*(p)) * (n); \
  DECBYTES (deleten_bytes); \
  free ((p)); \
  (p) = 0; \
} while (0)

#define NEW(p) NEWN((p),1)
#define DELETE(p) DELETEN((p),1)

#define ENLARGE(p,s) \
do { \
  unsigned enlarge_new_size = (s) ? 2 * (s) : 1; \
  REALLOCN ((p), (s), enlarge_new_size); \
  (s) = enlarge_new_size; \
} while (0)

#define TRUE ((BDD*) 1)
#define FALSE ((BDD*) 0)
#define SIGN(p) ((unsigned)(1lu&(unsigned long)(p)))
#define NOT(p) ((BDD*)(1lu^(unsigned long)(p)))
#define STRIP(p) ((BDD*)(~1lu&(unsigned long)(p)))
#define CONST(p) (1lu >= (unsigned long)(p))

#define AVERAGE(a,b) ((b) ? (a)/(double)(b) : 0.0)
#define PERCENTAGE(a,b) (100.0 * AVERAGE ((a), (b)))

#define SWAPBDD(a,b) \
  do { \
    BDD * swapbdd_tmp = (a); \
    (a) = (b); \
    (b) = swapbdd_tmp; \
  } while (0)

static void
die (const char *fmt, ...)
{
  va_list ap;

  fputs ("*** aigtrav: ", stderr);
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static void *
malloc4aiger (void *dummy, size_t bytes)
{
  char *res;
  (void) dummy;
  NEWN (res, bytes);
  return res;
}

static void
free4aiger (void *dummy, void *ptr, size_t bytes)
{
  char *char_ptr = ptr;
  DELETEN (char_ptr, bytes);
  (void) dummy;
}

static void
msg (const char *fmt, ...)
{
  int mtmsg = !strcmp (fmt, "");
  va_list ap;

  if (!verbose)
    return;

  if (mtmsg && last_line_was_empty)
    return;

  fputs ("[aigtrav]", stderr);
  if (mtmsg)
    {
      last_line_was_empty = 1;
    }
  else
    {
      fputc (' ', stderr);
      va_start (ap, fmt);
      vfprintf (stderr, fmt, ap);
      va_end (ap);
    }

  fputc ('\n', stderr);
  fflush (stderr);
  fflush (stdout);

  last_line_was_empty = !strcmp (fmt, "");
}

static unsigned
bdd2idx (BDD * bdd)
{
  unsigned sign, res;
  BDD *stripped;

  if (bdd == FALSE)
    return 0;

  if (bdd == TRUE)
    return 1;

  sign = SIGN (bdd);
  stripped = STRIP (bdd);

  res = 2 * stripped->idx;
  res += sign;

  return res;
}

static unsigned
hash (unsigned var, BDD * low, BDD * high)
{
  unsigned res = 0;

  res += 2112549121u * var;
  res += 2222905691u * bdd2idx (low);
  res += 2333960159u * bdd2idx (high);

  return res;
}

static size_t
mb_cache (void)
{
  size_t res;
  res = size_cache * sizeof cache[0];
  return res;
}

static void
enlarge_cache (void)
{
  Line *new_cache, *line, *eoc;
  unsigned new_size, h;

  if (size_cache)
    {
      if (mb_cache () >= (1u << 27))			/*>=128MB*/
	new_size = size_cache + (size_cache + 1) / 2;	/*+50%*/
      else
	new_size = 2 * size_cache;			/*+100%*/
    }
  else 
    new_size = 1;

  if (new_size > size_cache_limit)
    return ;

  msg ("increasing cache from %u to %u by %.0f%%", 
       size_cache, new_size, PERCENTAGE (new_size - size_cache, size_cache));

  NEWN (new_cache, new_size);

  count_cache = 0;
  eoc = cache + size_cache;
  for (line = cache; line < eoc; line++)
    {
      if (!line->op)
	continue;

      h = hash (0, line->a, line->b);
      h &= new_size - 1;

      if (!new_cache[h].op)
	{
	  assert (count_cache < new_size);
	  count_cache++;
	}

      new_cache[h] = *line;
    }

  DELETEN (cache, size_cache);

  cache = new_cache;
  size_cache = new_size;
}

static size_t
mb_bdds (void)
{
  size_t res;
  res = (num_bdds + num_recycled_bdds) * sizeof (BDD);
  res += size_bdds * sizeof (bdds[0]);
  return res;
}

static void
logmb (void)
{
  msg ("allocated %.0f MB, nodes %.0f MB (%.0f%%), cache %.0f MB (%.0f%%)",
       current_bytes / (double)(1<<20),
       mb_bdds () / (double)(1<<20), 
         PERCENTAGE (mb_bdds (), current_bytes),
       mb_cache () / (double)(1<<20), 
         PERCENTAGE (mb_cache (), current_bytes));
}

static void
enlarge_bdds (void)
{
  BDD *p, *next, * first;
  unsigned new_size, i, h;

  msg ("");

  if (size_bdds)
    {
      if (mb_bdds () >= (1u << 29))			/*>=512MB*/
	new_size = size_bdds + (size_bdds + 1) / 2;	/*+50%*/
      else
	new_size = 2 * size_bdds;			/*+100%*/
    }
  else
    new_size = 1;

  msg ("increasing table from %u to %u by %.0f%%", 
       size_bdds, new_size, PERCENTAGE (new_size - size_bdds, size_bdds));

  first = 0;

  for (i = 0; i < size_bdds; i++)
    {
      for (p = bdds[i]; p; p = next)
	{
	  next = p->next;
          p->next = first;
	  first = p;
	}
    }

  REALLOCN (bdds, size_bdds, new_size);
  CLRN (bdds, new_size);

  for (p = first; p; p = next)
    {
      next = p->next;
      h = hash (p->var, p->low, p->high);
      h &= new_size - 1;
      p->next = bdds[h];
      bdds[h] = p;
    }

  size_bdds = new_size;

  enlarge_cache ();

  if (verbose)
    logmb ();

  msg ("");
}

static BDD *
inc_bdd (BDD * bdd)
{
  BDD *stripped;

  stripped = STRIP (bdd);
  if (stripped)
    {
      stripped->ref++;
      assert (stripped->ref);
    }

  return bdd;
}

static void
recycle_bdd (BDD * bdd)
{
  unsigned h;
  BDD **p;

  bdd = STRIP (bdd);
  if (!bdd)
    return;

  assert (bdd->ref);
  bdd->ref--;
  if (bdd->ref)
    return;

  assert (num_bdds);

  h = hash (bdd->var, bdd->low, bdd->high);
  h &= size_bdds - 1;

  for (p = bdds + h; *p != bdd; p = &(*p)->next)
    assert (*p);

  *p = bdd->next;
  bdd->next = recycled_bdds;
  recycled_bdds = bdd;
  num_bdds--;
  num_recycled_bdds++;

  recycle_bdd (bdd->low);
  recycle_bdd (bdd->high);
}

static unsigned
bdd2ref (BDD * bdd)
{
  bdd = STRIP (bdd);
  if (!bdd)
    return UINT_MAX;

  return bdd->ref;
}

static void
recycle_cache (void)
{
  Line *p, *eoc = cache + size_cache;
  unsigned recycled = 0;

  msg ("%u cache lines out of %u (%.0f%%)",
       count_cache, size_cache, PERCENTAGE (count_cache, size_cache));

  for (p = cache; p < eoc; p++)
    {
      if (!p->op)
	continue;

      if (bdd2ref (p->a) && bdd2ref (p->b) && bdd2ref (p->r))
	continue;

      p->op = 0;
      recycled++;
      assert (count_cache);
      count_cache--;
    }

  if (recycled)
    msg ("recycled %u cache lines (%.0f%%)",
	 recycled, PERCENTAGE (recycled, size_cache));
}

static void
recycle_dead_bdds (unsigned target_remaining)
{
  unsigned old_num_bdds, recycled, flushed, removed;
  unsigned i, j, target_removed;
  BDD *b;

  old_num_bdds = num_bdds;

  msg ("");
  msg ("trying to shrink death row from %u to %u nodes",
       num_dead_bdds, target_remaining);

  j = 0;

  for (i = 0; i < num_dead_bdds; i++)
    {
      b = dead_bdds[i];
      if (!b->ref)
	dead_bdds[j++] = b;
    }
  flushed = num_dead_bdds - j;
  if (flushed)
    msg ("flushed %u resurrected bdd roots", flushed);
  num_dead_bdds = j;

  /* Then remove duplicates, keep latest added.
   */
  i = j = num_dead_bdds;
  while (i)
    {
      b = dead_bdds[--i];
      if (!b->mark)
	{
	  b->mark = 1;
	  dead_bdds[--j] = b;
	}
    }

  assert (!i);
  while (j < num_dead_bdds)
    {
      b = dead_bdds[j++];
      assert (b->mark);
      b->mark = 0;
      dead_bdds[i++] = b;
    }
  removed = num_dead_bdds - i;
  if (removed)
    msg ("removed %u duplicates among dead bdd roots", removed);
  num_dead_bdds = i;

  if (num_dead_bdds > target_remaining)
    {
      target_removed = num_dead_bdds - target_remaining;
      msg ("recycling recursively %u dead bdd roots", target_removed);

      for (i = 0; i < target_removed; i++)
	{
	  b = dead_bdds[i];
	  assert (b);
	  assert (!SIGN (b));
	  assert (!b->ref);
	  b->ref = 1;
	  recycle_bdd (dead_bdds[i]);
	}

      for (j = target_removed; j < num_dead_bdds; j++)
	dead_bdds[j]->ref = 1;

      recycle_cache ();

      for (j = target_removed; j < num_dead_bdds; j++)
	dead_bdds[j]->ref = 0;

      j = 0;
      while (i < num_dead_bdds)
	dead_bdds[j++] = dead_bdds[i++];
      num_dead_bdds = j;

      assert (num_dead_bdds == target_remaining);
    }

  assert (old_num_bdds >= num_bdds);
  recycled = old_num_bdds - num_bdds;
  if (recycled)
    msg ("recycled %u bdd nodes (%.0f%%)",
	 recycled, PERCENTAGE (recycled, old_num_bdds));

  if (!flushed && !removed && !recycled)
    msg ("WARNING: GC did not flush, nor remove, nor recycle any nodes");

  msg ("%u dead nodes remaining", num_dead_bdds);
  msg ("");
}

static void
kill_node (BDD * bdd)
{
  assert (!bdd->ref);

  if (num_dead_bdds == size_dead_bdds)
    ENLARGE (dead_bdds, size_dead_bdds);

  dead_bdds[num_dead_bdds++] = bdd;

  if (num_dead_bdds >= num_dead_bdds_limit)
    {
      recycle_dead_bdds (num_dead_bdds / 2);

      if (3*num_dead_bdds_limit/4 <= num_dead_bdds)
	{
	  num_dead_bdds_limit = 11*num_dead_bdds_limit/10;
	  msg ("increasing dead bdd limit to %u", num_dead_bdds_limit);
	  msg ("");
	}
    }
}

static void
dec_bdd (BDD * bdd)
{
  BDD *stripped;

  stripped = STRIP (bdd);
  if (stripped)
    {
      assert (stripped->ref);
      stripped->ref--;
      if (!stripped->ref)
	kill_node (stripped);
    }
}

static unsigned
var2current (unsigned var)
{
  return !(var & 1);
}

static unsigned
bdd2var (BDD * bdd)
{
  assert (!CONST (bdd));
  return STRIP (bdd)->var;
}

static BDD *
new_bdd (unsigned var, BDD * low, BDD * high)
{
  unsigned sign, h;
  BDD **p, *b;

  assert (CONST (low) || var > bdd2var (low));
  assert (CONST (high) || var > bdd2var (high));

  if (low == high)
    return inc_bdd (low);

  if (num_bdds == size_bdds)
    enlarge_bdds ();

  sign = SIGN (low);
  if (sign)
    {
      low = NOT (low);
      high = NOT (high);
    }

  h = hash (var, low, high);
  h &= size_bdds - 1;

  unique_table_lookups++;

  for (p = bdds + h;
       (b = *p) && !(b->var == var && b->low == low && b->high == high);
       p = &b->next)
    unique_table_visits++;

  if (!b)
    {
      if (recycled_bdds)
	{
	  b = recycled_bdds;
	  recycled_bdds = b->next;
	  b->next = 0;
	  b->ref = 0;

	  assert (!b->mark);

	  assert (num_recycled_bdds);
	  num_recycled_bdds--;
	}
      else
	NEW (b);

      b->idx = idx_bdds++;
      b->var = var;
      b->low = inc_bdd (low);
      b->high = inc_bdd (high);

      num_bdds++;

      *p = b;
    }

  b->ref++;

  if (sign)
    b = NOT (b);

  return b;
}

static BDD *
bdd2low (BDD * bdd)
{
  BDD *stripped = STRIP (bdd), *res;
  unsigned sign = SIGN (bdd);
  res = stripped->low;
  if (sign)
    res = NOT (res);
  return res;
}

static BDD *
bdd2high (BDD * bdd)
{
  BDD *stripped = STRIP (bdd), *res;
  unsigned sign = SIGN (bdd);
  res = stripped->high;
  if (sign)
    res = NOT (res);
  return res;
}

static BDD *
apply (Op op, BDD * a, BDD * b)
{
  BDD *res, *low, *high, *alow, *ahigh, *blow, *bhigh;
  unsigned h, var;
  Line *line;

  assert (apply_steps <= apply_steps_limit);

  apply_steps++;

  if (verbose > 1)
    {
      if (op == AND)
	rands++;
      else if (op == EY)
	reys++;
      else
	{
	  assert (op == XOR);
	  rxors++;
	}
    }

  if (apply_steps > apply_steps_limit)
    {
      msg ("limit of %llu apply steps reached", apply_steps_limit);
      return FALSE;
    }

  if (op == XOR)
    {
      if (a == FALSE)
	return inc_bdd (b);

      if (a == TRUE)
	return NOT (inc_bdd (b));

      if (b == FALSE)
	return inc_bdd (a);

      if (b == TRUE)
	return NOT (inc_bdd (a));

      if (a == b)
	return FALSE;

      if (a == NOT (b))
	return TRUE;
    }
  else
    {
      assert (op == AND || op == EX || op == EY);

      if (a == FALSE || b == FALSE || a == NOT (b))
	return FALSE;

      if (op == AND)
	{
	  if (a == TRUE || a == b)
	    return inc_bdd (b);

	  if (b == TRUE)
	    return inc_bdd (a);
	}
      else
	{
	  if (b == TRUE)
	    {
	      if (a == TRUE)
		return TRUE;

	      SWAPBDD (a, b);
	    }

	  if (a == TRUE)
	    a = b;
	}
    }

  if (bdd2var (a) > bdd2var (b))
    SWAPBDD (a, b);

  if (bdd2var (a) == bdd2var (b) && bdd2idx (a) > bdd2idx (b))
    SWAPBDD (a, b);

  h = hash (0, a, b);
  line = cache + (h & (size_cache - 1));
  cache_lookups++;
  if (line->op == op && line->a == a && line->b == b)
    {
      cache_hits++;
      return inc_bdd (line->r);
    }

  var = bdd2var (b);

  if (var == bdd2var (a))
    {
      alow = bdd2low (a);
      ahigh = bdd2high (a);
    }
  else
    alow = ahigh = a;

  blow = bdd2low (b);
  bhigh = bdd2high (b);

  low = apply (op, alow, blow);

  if (apply_steps > apply_steps_limit)
    {
      assert (low == FALSE);
      return FALSE;
    }

  high = apply (op, ahigh, bhigh);

  if (apply_steps > apply_steps_limit)
    {
      assert (high == FALSE);
      dec_bdd (low);
      return FALSE;
    }

  if (op == EY)
    {
      if (var2current (var))
	{
	  res = NOT (apply (AND, NOT (low), NOT (high)));

	  if (apply_steps > apply_steps_limit)
	    {
	      assert (res == FALSE);
	      dec_bdd (low);
	      dec_bdd (high);
	      return FALSE;
	    }
	}
      else
	{
	  assert (var > 0);
	  res = new_bdd (var - 1, low, high);
	}
    }
  else if (op == EX)
    {
      if (!var2current (var))
	{
	  res = NOT (apply (AND, NOT (low), NOT (high)));

	  if (apply_steps > apply_steps_limit)
	    {
	      assert (res == FALSE);
	      dec_bdd (low);
	      dec_bdd (high);
	      return FALSE;
	    }
	}
      else
	res = new_bdd (var + 1, low, high);
    }
  else
    res = new_bdd (var, low, high);

  dec_bdd (high);
  dec_bdd (low);

  assert (h == hash (0, a, b));

  line = cache + (h & (size_cache - 1));

  if (!line->op)
    {
      assert (count_cache < size_cache);
      count_cache++;
    }

  line->op = op;
  line->a = a;
  line->b = b;
  line->r = res;

  return res;
}

static unsigned
mark_bdd (BDD * a, unsigned new_mark)
{
  unsigned res;

  a = STRIP (a);
  if (!a)
    return 0;

  if (a->mark == new_mark)
    return 0;

  a->mark = new_mark;

  res = 1;
  res += mark_bdd (a->low, new_mark);
  res += mark_bdd (a->high, new_mark);

  return res;
}

static unsigned
count_bdd (BDD * a)
{
  unsigned res = mark_bdd (a, 1);
#ifndef NDEBUG
  unsigned tmp = 
#endif
  mark_bdd (a, 0);
  assert (res == tmp);
  return res;
}

static BDD *
var2bdd (unsigned var)
{
  return new_bdd (var, FALSE, TRUE);
}

static int
need_parenthesis (Op outer, Op inner)
{
  return 1;
  if (outer < inner)
    return 0;

  if (outer > inner)
    return 1;

  return inner == ITE;
}

static void
open_parenthesis (Op outer, Op inner)
{
  if (need_parenthesis (outer, inner))
    printf ("( ");
}

static void
close_parenthesis (Op outer, Op inner)
{
  if (need_parenthesis (outer, inner))
    printf (" )");
}

static void printaux (BDD *, Op);

static void
print_signed_var_and_op (unsigned sign, unsigned var, char op)
{
  printf ("%sv%u %c ", sign ? "!" : "", var, op);
}

static void
print_and (Op outer, unsigned sign, unsigned a, BDD * b)
{
  open_parenthesis (outer, AND);
  print_signed_var_and_op (sign, a, '&');
  printaux (b, AND);
  close_parenthesis (outer, AND);
}

static void
print_or (Op outer, unsigned sign, unsigned a, BDD * b)
{
  open_parenthesis (outer, OR);
  print_signed_var_and_op (sign, a, '|');
  printaux (b, OR);
  close_parenthesis (outer, OR);
}

static void
print_xor (Op outer, unsigned a, BDD * b)
{
  open_parenthesis (outer, XOR);
  print_signed_var_and_op (0, a, '^');
  printaux (b, XOR);
  close_parenthesis (outer, XOR);
}

static void
print_ite (Op outer, unsigned var, BDD * high, BDD * low)
{
  open_parenthesis (outer, ITE);
  printf ("v%u ? ", var);
  printaux (high, ITE);
  fputs (" : ", stdout);
  printaux (low, ITE);
  close_parenthesis (outer, ITE);
}

static void
printaux (BDD * bdd, Op outer)
{
  BDD *stripped;
  unsigned sign;

  if (bdd == FALSE)
    fputc ('0', stdout);
  else if (bdd == TRUE)
    fputc ('1', stdout);
  else
    {
      sign = SIGN (bdd);
      stripped = STRIP (bdd);

      if (stripped->low == FALSE && stripped->high == TRUE)
	{
	  if (sign)
	    fputc ('!', stdout);

	  printf ("v%u", stripped->var);
	}
      else if (stripped->low == FALSE)
	{
	  /* var ? high : 0 == var & high
	   */
	  if (sign)
	    print_or (outer, 1, stripped->var, NOT (stripped->high));
	  else
	    print_and (outer, 0, stripped->var, stripped->high);
	}
      else if (stripped->low == TRUE)
	{
	  /* var ? high : 1 == !var | high
	   */
	  if (sign)
	    print_and (outer, 0, stripped->var, NOT (stripped->high));
	  else
	    print_or (outer, 1, stripped->var, stripped->high);
	}
      else if (stripped->high == FALSE)
	{
	  /* var ? 0 : low == !var & low
	   */
	  if (sign)
	    print_or (outer, 0, stripped->var, NOT (stripped->low));
	  else
	    print_and (outer, 1, stripped->var, stripped->low);
	}
      else if (stripped->high == TRUE)
	{
	  /* var ? 1 : low == var | !low
	   */
	  if (sign)
	    print_and (outer, 1, stripped->var, NOT (stripped->low));
	  else
	    print_or (outer, 0, stripped->var, stripped->low);
	}
      else if (stripped->low == NOT (stripped->high))
	{
	  /* var ? !low : low  == var ^ low
	   */
	  if (sign)
	    print_xor (outer, stripped->var, NOT (stripped->low));
	  else
	    print_xor (outer, stripped->var, stripped->low);
	}
      else
	{
	  // assert (STRIP (stripped->low) != STRIP (stripped->high));
	  if (sign)
	    print_ite (outer,
		       stripped->var,
		       NOT (stripped->high), NOT (stripped->low));
	  else
	    print_ite (outer, stripped->var, stripped->high, stripped->low);
	}
    }
}

void
print (BDD * bdd)
{
  printaux (bdd, EMPTY);
  fputc ('\n', stdout);
}

static void
release_bdds (void)
{
  BDD *p, *next;
  unsigned i;

  for (i = 0; i < size_bdds; i++)
    {
      for (p = bdds[i]; p; p = next)
	{
	  next = p->next;
	  DELETE (p);
	}
    }

  DELETEN (bdds, size_bdds);

  for (p = recycled_bdds; p; p = next)
    {
      next = p->next;
      DELETE (p);
    }
}

static void
push_unsigned_stack (unsigned lit)
{
  if (size_unsigned_stack == count_unsigned_stack)
    ENLARGE (unsigned_stack, size_unsigned_stack);

  unsigned_stack[count_unsigned_stack++] = lit;
}

static unsigned
pop_unsigned_stack (void)
{
  assert (count_unsigned_stack);
  return unsigned_stack[--count_unsigned_stack];
}

static void
map (void)
{
  aiger_symbol * symbol;
  unsigned lit, l0, l1;
  aiger_and * and;
  AIG * aig;

  NEWN (aigs, aiger_model->maxvar + 1);

  for (lit = 1; lit <= aiger_model->maxvar; lit++)
    {
      aig = aigs + lit;

      if ((symbol = aiger_is_input (aiger_model, 2 * lit)))
	aig->input = 1;
      else if ((symbol = aiger_is_latch (aiger_model, 2 * lit)))
	{
	  aig->latch = 1;
	  aig->rhs[0] = symbol->next;
	  aig->rhs[1] = symbol->reset;
	}
      else
	{
	  aig->and = 1;
	  and = aiger_is_and (aiger_model, 2 * lit);
	  assert (and); /* reencoded! */
	  aig->rhs[0] = and->rhs0;		/* signed! */
	  aig->rhs[1] = and->rhs1;		/* signed! */
	  l0 = aigs[aig->rhs[0]/2].level;
	  l1 = aigs[aig->rhs[1]/2].level;
	  aig->level = ((l0 > l1) ? l0 : l1) + 1;
	}
    }
}

static unsigned
bad (void)
{
  unsigned res;
  if (aiger_model->num_bad) {
    assert (aiger_model->num_bad == 1);
    res = aiger_model->bad[0].lit/2;
  } else {
    assert (aiger_model->num_outputs == 1);
    res = aiger_model->outputs[0].lit/2;
  }
  return res;
}

static void
coi (void)
{
  unsigned lit;
  AIG * aig;

  assert (!count_unsigned_stack);
  lit = bad ();
  if (lit)
    push_unsigned_stack (lit);

  while (count_unsigned_stack)
    {
      lit = pop_unsigned_stack ();
CONTINUE_WITHOUT_POP:
      if (!lit)
	continue;

      aig = aigs + lit;

      if (aig->coi)
	continue;

      aig->coi = 1;

      if (aig->input)
	stats.coi.inputs++;
      else if (aig->latch)
	{
	  stats.coi.latches++;
	  lit = aig->rhs[0] / 2;
	  goto CONTINUE_WITHOUT_POP;
	}
      else
	{
	  assert (aig->and);
	  stats.coi.aigs++;

	  lit = aig->rhs[1] / 2;
	  push_unsigned_stack (lit);

	  lit = aig->rhs[0] / 2;
	  goto CONTINUE_WITHOUT_POP;
	}
    }

  msg ("coi: %u latches (%.0f%%), %u inputs (%.0f%%), %u aigs (%.0f%%)",
       stats.coi.latches,
         PERCENTAGE (stats.coi.latches, aiger_model->num_latches),
       stats.coi.inputs,
         PERCENTAGE (stats.coi.inputs, aiger_model->num_inputs),
       stats.coi.aigs,
         PERCENTAGE (stats.coi.aigs, aiger_model->num_ands));
}

static void
abstract (unsigned latch_limit)
{
  unsigned next, lit, latch_count, rhs, i;
  AIG * aig, * child;

  stats.abstraction.count++;

  assert (!count_unsigned_stack);
  lit = bad ();
  if (lit)
    {
      assert (!aigs[lit].mark);
      aigs[lit].mark = 1;
      push_unsigned_stack (lit);
    }

  next = 0;
  latch_count = 0;
  while (next != count_unsigned_stack)
    {
      lit = unsigned_stack[next++];
      assert (aigs[lit].mark);
      assert (lit);

      aig = aigs + lit;
      assert (aig->coi);

      if (aig->latch)
	{
	  if (latch_count >= latch_limit)
	    {
	      next--;
	      break;
	    }

	  latch_count++;
	  rhs = aig->rhs[0]/2;
	  child = aigs + rhs;
	  if (rhs && !child->mark)
	    {
	      child->mark = 1;
	      push_unsigned_stack (rhs);
	    }
	}
      else if (aig->and)
	{
	  for (i = 0; i <= 1; i++)
	    {
	      rhs = aig->rhs[i]/2;
	      child = aigs + rhs;
	      if (rhs && !child->mark)
		{
		  child->mark = 1;
		  push_unsigned_stack (rhs);
		}
	    }
	}
      else
	assert (aig->input);		/* and nothing to be done */
    }

  for (i = 0; i < next; i++)
    {
      aig = aigs + unsigned_stack[i];
      assert (aig->coi);

      assert (aig->mark);
      aig->mark = 0;

      assert (!aig->included);
      aig->included = 1;

      if (aig->input)
	stats.abstraction.inputs++;
      else if (aig->latch)
	{
	  assert (!aig->freed);
	  stats.abstraction.latches++;
	}
      else
	stats.abstraction.aigs++;
    }

  while (i < count_unsigned_stack)
    {
      aig = aigs + unsigned_stack[i++];
      assert (aig->coi);

      assert (aig->mark);
      aig->mark = 0;

      assert (!aig->included);
      aig->included = 1;

      if (aig->latch || aig->and)
	{
	  assert (!aig->freed);
	  aig->freed = 1;
	  stats.abstraction.freed++;
	}
    }

  count_unsigned_stack = 0;

  msg ("");
  msg ("     abs[%u] %u latches (%.0f%%), %u freed (%.0f%%), %u limit",
       stats.abstraction.count,
       stats.abstraction.latches,
         PERCENTAGE (stats.abstraction.latches, stats.coi.latches),
       stats.abstraction.freed,
         PERCENTAGE (stats.abstraction.freed, stats.coi.latches),
       latch_limit);

  msg ("     abs[%u] %u inputs (%.0f%%), %u aigs (%.0f%%)",
       stats.abstraction.count,
       stats.abstraction.inputs,
         PERCENTAGE (stats.abstraction.inputs, stats.coi.inputs),
       stats.abstraction.aigs,
         PERCENTAGE (stats.abstraction.aigs, stats.coi.aigs));
}

static void
unmark (void)
{
  AIG * aig, * child;
  unsigned lit, i;

  assert (!count_unsigned_stack);
  lit = bad ();
  if (lit)
    {
      aig = aigs + lit;
      assert (aig->mark);
      aig->mark = 0;
      push_unsigned_stack (lit);
    }

  while (count_unsigned_stack)
    {
      lit = pop_unsigned_stack ();
      if (!lit)
	continue;

      aig = aigs + lit;

      assert (!aig->mark);

      if (aig->latch)
	{
	  lit = aig->rhs[0] / 2;
	  child = aigs + lit;
	  if (lit && child->mark)
	    {
	      child->mark = 0;
	      push_unsigned_stack (lit);
	    }
	}
      else if (aig->and)
	{
	  for (i = 0; i <= 1; i++)
	    {
	      lit = aig->rhs[i] / 2;
	      child = aigs + lit;
	      if (lit && child->mark)
		{
		  child->mark = 0;
		  push_unsigned_stack (lit);
		}
	    }
	}
    }

#ifndef NDEBUG
  for (aig = aigs + 1; aig <= aigs + aiger_model->maxvar; aig++)
    assert (!aig->mark);
#endif
}

static void
order (void)
{
  unsigned lit, var, i, rhs[2], first;
  AIG * aig, * child[2], * tmp;

  msg ("");

#ifndef NDEBUG
  for (aig = aigs + 1; aig <= aigs + aiger_model->maxvar; aig++)
    assert (!aig->mark);
#endif

  assert (!count_unsigned_stack);
  lit = bad ();
  if (lit)
    {
      aig = aigs + lit;
      assert (!aig->mark);
      aig->mark = 1;
      push_unsigned_stack (lit);
    }

  var = 0;

  while (count_unsigned_stack)
    {
      lit = pop_unsigned_stack ();
      if (!lit)
	continue;

      aig = aigs + lit;

      assert (aig->coi);
      assert (aig->included);
      assert (aig->mark);

      if (aig->input || aig->freed)
	{
	  aig->var = var++;
	}
      else if (aig->latch)
	{
	  assert (!aig->freed);

	  aig->var = var++;
	  lit = aig->rhs[0] / 2;
	  tmp = aigs + lit;
	  if (lit && tmp->included && !tmp->mark)
	    {
	      tmp->mark = 1;
	      push_unsigned_stack (lit);
	    }
	}
      else
	{
	  assert (aig->and);
	  assert (!aig->freed);

	  for (i = 0; i <= 1; i++)
	    {
	      rhs[i] = aig->rhs[i] / 2;
	      tmp = aigs + rhs[i];
	      assert (!rhs[i] || tmp->included);
	      child[i] = (rhs[i] && !tmp->mark) ? tmp : 0;
	    }

	  if (child[0] && child[1])
	    {
	      first = (child[1]->level > child[0]->level);

	      child[!first]->mark = 1;
	      push_unsigned_stack (rhs[!first]);

	      child[first]->mark = 1;
	      push_unsigned_stack (rhs[first]);
	    }
	  else
	    {
	      for (i = 0; i <= 1; i++)
		{
		  if (child[i])
		    {
		      child[i]->mark = 1;
		      push_unsigned_stack (rhs[i]);
		    }
		}
	    }
	}
    }

  unmark ();

  msg ("ordered %u variables", var);
  msg ("");
}

static void
inc_aig_ref (unsigned idx)
{
  assert (idx <= aiger_model->maxvar);
  if (idx)
    aigs[idx].ref++;
}

static BDD *
lit2bdd (unsigned lit, int deref)
{
  AIG * aig;
  BDD * res;

  if (lit == 0)
    return FALSE;

  if (lit == 1)
    return TRUE;

  assert (lit < 2 * (aiger_model->maxvar + 1));

  aig = aigs + (lit/2);

  assert (aig->ref);
  assert (aig->valid);
  assert (aig->included);
  assert (aig->coi);

  res = aig->bdd;
  if (lit & 1)
    res = NOT (res);

  if (deref)
    {
      aig->ref--;
      if (aig->and && !aig->freed && !aig->ref)
	{
	  aig->valid = 0;
	  dec_bdd (aig->bdd);
	}
    }

  return res;
}

static void
logop (const char * opname,
       unsigned opsteps, unsigned long long delta_apply_steps,
       BDD * a, BDD * b, BDD * r)
{
  unsigned long long max;
  unsigned c, d; 

  if (apply_steps > apply_steps_limit)
    return;

  c = count_bdd (a);
  d = count_bdd (b);
  max = c * (unsigned long long) d;

  msg ("%8s[%u] #%u (%u, %u) = %u (%llu steps = %.0f%% of %llu)",
       opname, stats.abstraction.count, opsteps, c, d, count_bdd (r),
       delta_apply_steps, PERCENTAGE (delta_apply_steps, max), max);
}

static BDD *
and (BDD * a, BDD * b)
{
  unsigned long long old_apply_steps = apply_steps;
  BDD * res = apply (AND, a, b);
  if (verbose > 1)
    logop ("and", ands++, apply_steps - old_apply_steps, a, b, res);
  return res;
}

static BDD *
xor (BDD * a, BDD * b)
{
  unsigned long long old_apply_steps = apply_steps;
  BDD * res = apply (XOR, a, b);
  if (verbose > 1)
    logop ("xor", xors++, apply_steps - old_apply_steps, a, b, res);
  return res;
}

static BDD *
or (BDD * a, BDD * b)
{
  unsigned long long old_apply_steps = apply_steps;
  BDD * res = NOT (apply (AND, NOT (a), NOT (b)));
  if (verbose > 1)
    logop ("or", ors++, apply_steps - old_apply_steps, a, b, res);
  return res;
}

static BDD *
ey (BDD * a, BDD * b)
{
  unsigned long long old_apply_steps = apply_steps;
  BDD * res = apply (EY, a, b);
  if (verbose > 1)
    logop ("ey", eys++, apply_steps - old_apply_steps, a, b, res);
  return res;
}

static void
build (void)
{
  AIG * aig, * eoa;

  msg ("");
  inc_aig_ref (bad ());

  eoa = aigs + aiger_model->maxvar;
  for (aig = aigs + 1; apply_steps <= apply_steps_limit && aig <= eoa; aig++)
    {
      if (!aig->included)
	continue;

      if (aig->freed)
	continue;

      if (aig->and)
	{
	  inc_aig_ref (aig->rhs[0]/2);
	  inc_aig_ref (aig->rhs[1]/2);
	}
      else if (aig->latch)
	inc_aig_ref (aig->rhs[0]/2);
    }

  for (aig = aigs + 1; apply_steps <= apply_steps_limit && aig <= eoa; aig++)
    {
      if (!aig->included)
	continue;

      assert (aig->coi);
      aig->valid = 1;

      if (aig->and && !aig->freed)
	aig->bdd = and (lit2bdd (aig->rhs[0], 1), lit2bdd (aig->rhs[1], 1));
      else
	aig->bdd = var2bdd (2 * aig->var);
    }

  for (aig = aigs + 1; apply_steps <= apply_steps_limit && aig <= eoa; aig++)
    {
      if (!aig->included)
	continue;

      if (!aig->latch)
	continue;

      if (aig->freed)
	continue;

      aig->next = inc_bdd (lit2bdd (aig->rhs[0], 1));
    }

  // recycle_dead_bdds (0);

  msg ("   model[%u] %u nodes", stats.abstraction.count, num_bdds);
  msg ("");
}

static int
traverse (void)
{
  BDD * next, * current, * reached, * prev, * tmp, * assign;
  BDD * init, * trans, * target;
  AIG * aig, * eoa;
  unsigned steps;
  int res;

  init = TRUE;
  trans = TRUE;
  
  eoa = aigs + aiger_model->maxvar;
  for (aig = aigs + 1; apply_steps <= apply_steps_limit && aig <= eoa; aig++)
    {
      if (!aig->included)
	continue;

      assert (aig->coi);

      if (!aig->latch)
	continue;

      if (aig->freed)
	continue;

      if (aig->rhs[1] == aiger_false || aig->rhs[1] == aiger_true)
	{
	  current = var2bdd (2 * aig->var);
	  if (aig->rhs[1] == aiger_false)
	    current = NOT (current);
	  else
	    assert (aig->rhs[1] == aiger_true);
	  tmp = and (init, current);
	  dec_bdd (init);
	  dec_bdd (current);
	  init = tmp;
	}

      if (apply_steps > apply_steps_limit)
	break;

      next = var2bdd (2 * aig->var + 1);
      assign = xor (NOT (next), aig->next);
      dec_bdd (assign);
      tmp = and (trans, assign);
      dec_bdd (trans);
      dec_bdd (next);
      trans = tmp;
    }

  target = lit2bdd (aiger_model->outputs[0].lit, 0);

  msg ("    init[%u] %u nodes", stats.abstraction.count, count_bdd (init));
  msg ("   trans[%u] %u nodes", stats.abstraction.count, count_bdd (trans));
  msg ("  target[%u] %u nodes", stats.abstraction.count, count_bdd (target));

  steps = 0;
  prev = inc_bdd (init);
  reached = inc_bdd (init);
  res = -1;

  while (apply_steps <= apply_steps_limit)
    {
      msg ("frontier[%u] %u nodes",
	   stats.abstraction.count, count_bdd (prev));

      tmp = and (prev, target);
      if (tmp)
	{ 
	  msg ("     hit[%u] target", stats.abstraction.count);
	  dec_bdd (tmp);
	  res = 1;
	  break;
	}

      dec_bdd (tmp);		/* redundant */

      steps++;

      msg ("   image[%u] %u steps", stats.abstraction.count, steps);

      next = ey (trans, prev);
      dec_bdd (prev);

      prev = and (NOT (reached), next);
      dec_bdd (next);

      if (!prev)
        {
	  msg ("fixpoint[%u] reached", stats.abstraction.count);
	  break;
	}

      tmp = or (reached, prev);
      dec_bdd (reached);
      reached = tmp;

      msg (" reached[%u] %u nodes",
	   stats.abstraction.count, count_bdd (reached));
    }

  if (res < 0) {
    print(reached);
    unsigned lit;
    AIG *aig;
    for (lit = 1; lit <= aiger_model->maxvar; lit++) {
      aig = aigs + lit;
      if (!aig->latch) continue;
      if (!aig->bdd) continue;
      printf("%d -> %d\n", aig->bdd->var, 2*lit);
    }
  }

  dec_bdd (prev);
  dec_bdd (trans);
  dec_bdd (init);

  dec_bdd (reached);

  msg ("");

  return res;
}

static void
reset (void)
{
  AIG * aig, * eoa;

  eoa = aigs + aiger_model->maxvar;
  for (aig = aigs + 1; aig <= eoa; aig++)
    {
      if (aig->included)
	{
	  assert (aig->coi);

	  if (aig->ref)
	    {
	      assert (aig->ref == 1);
	      assert ((aig - aigs) == (int) bad ());
	    }

	  if (aig->valid)
	    {
	      assert (aig->valid);
	      dec_bdd (aig->bdd);
	      if (aig->latch && !aig->freed)
		dec_bdd (aig->next);
	    }

	  aig->included = 0;
	  aig->freed = 0;
	  aig->valid = 0;
	  aig->ref = 0;
	}
      else
	assert (!aig->valid);
    }

  stats.abstraction.latches = 0;
  stats.abstraction.freed = 0;
  stats.abstraction.inputs = 0;
  stats.abstraction.aigs = 0;
}

int
main (int argc, char **argv)
{
  const char *input_name, *error_message;
  struct rusage rusage;
  unsigned latch_limit;
  int i, res, full = 0;

  input_name = 0;
  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "-h"))
	{
	  fprintf (stderr, 
	    "usage: aigtrav [-f][-l <limit>][-v][-h][<aiger-input-file>]\n");
	  exit (0);
	}
      else if (!strcmp (argv[i], "-f"))
	full = 1;
      else if (!strcmp (argv[i], "-l"))
	{
	  if (++i == argc)
	    die ("argument to '-l' missing");

	  apply_steps_limit = atoll (argv[i]);
	}
      else if (!strcmp (argv[i], "-v"))
	verbose++;
      else if (argv[i][0] == '-')
	die ("invalid command line option '%s'", argv[i]);
      else if (input_name)
	die ("multiple input files");
      else
	input_name = argv[i];
    }

  aiger_model = aiger_init_mem (0, malloc4aiger, free4aiger);

  msg ("$Id: aigtrav.c,v 1.38 2007-05-04 16:11:09 biere Exp $");
  msg ("parsing %s", input_name ? input_name : "<stdin>");

  if (input_name)
    error_message = aiger_open_and_read_from_file (aiger_model, input_name);
  else
    error_message = aiger_read_from_file (aiger_model, stdin);

  if (error_message)
    die (error_message);

  msg ("MILOA %d %d %d %d %d",
    aiger_model->maxvar,
    aiger_model->num_inputs,
    aiger_model->num_latches,
    aiger_model->num_outputs,
    aiger_model->num_ands);

  msg ("BCJF %d %d %d %d",
    aiger_model->num_bad,
    aiger_model->num_constraints,
    aiger_model->num_justice,
    aiger_model->num_fairness);

  if (aiger_model->num_constraints)
    die ("can not handle constraints");

  if (!aiger_model->num_outputs && !aiger_model->num_bad)
    die ("no output nor bad state property found");

  if (aiger_model->num_bad > 1)
    die ("can not handle multiple bad state properties");

  if (!aiger_model->num_bad && aiger_model->num_outputs > 1)
    die ("no bad state property but more than one output");

  if (aiger_model->num_justice)
    msg ("ignoring %d justice properties", aiger_model->num_justice);

  if (aiger_model->num_fairness)
    msg ("ignoring %d fairness constraints", aiger_model->num_fairness);

  aiger_reencode (aiger_model);

  msg ("");

  msg ("aig: %u latches, %u inputs, %u aigs",
       aiger_model->num_latches, 
       aiger_model->num_inputs,
       aiger_model->num_ands);

  map ();
  coi ();

  msg ("");

  latch_limit = full ? INT_MAX : 0;
  res = 0;

  while (apply_steps <= apply_steps_limit)
    {
      abstract (latch_limit);
      order ();
      build ();

      res = traverse ();
      reset ();

      if (res < 0)
	break;

      if (latch_limit >= stats.coi.latches)
	break;

      if (latch_limit)
	latch_limit += (latch_limit + 9) / 10;
      else
	latch_limit = 1;
    }

  msg ("");

  if (res < 0)
    {
      msg ("result: UNSATISFIABLE");
      printf ("0\n");
      fflush (stdout);
    }
  else if (res > 0)
    {
      msg ("result: SATISFIABLE");
      printf ("1\n");
      fflush (stdout);
    }
  else
    {
      msg ("result: UNKNOWN");
      printf ("x\n");
      fflush (stdout);
    }

  msg ("");
  msg ("%llu hits in %llu cache lookups (hit ratio %.0f%%)",
       cache_hits, cache_lookups, PERCENTAGE (cache_hits, cache_lookups));
  msg ("%llu unique table lookups (%.1f average visited)", 
       unique_table_lookups,
       AVERAGE (unique_table_visits, unique_table_lookups));
  msg ("%llu apply steps", apply_steps);
  msg ("");

  if (verbose > 1)
    {
      msg ("operations:  top level  apply steps");
      msg ("");
      msg ("    OR    %11u", ors);
      msg ("    EY    %11u %11u", eys, reys);
      msg ("   AND    %11u %11u", ands, rands);
      msg ("   XOR    %11u %11u", xors, rxors);
    }

  msg ("");
  msg ("%.0f MB max", max_bytes / (double) (1 << 20));

  DELETEN (cache, size_cache);
  DELETEN (dead_bdds, size_dead_bdds);

  release_bdds ();

  DELETEN (aigs, aiger_model->maxvar + 1);
  DELETEN (unsigned_stack, size_unsigned_stack);

  aiger_reset (aiger_model);

  getrusage (RUSAGE_SELF, &rusage);
  msg ("%.0f seconds",
       (rusage.ru_utime.tv_sec + 1e-6 * rusage.ru_utime.tv_usec) +
       (rusage.ru_stime.tv_sec + 1e-6 * rusage.ru_stime.tv_usec));

  assert (getenv ("LEAK") || !current_bytes);

  return 0;
}
